from __future__ import annotations

import argparse
import glob
import hashlib
import json
import os
import stat
import subprocess
import sys
import textwrap
import time
import typing as t

from slyp.checkers import check_file
from slyp.codes import CODE_MAP
from slyp.file_cache import HashedFile, PassingFileCache
from slyp.fixer import fix_file

DEFAULT_DISABLED_CODES: set[str] = {"W201", "W202", "W203"}


def main() -> None:
    parser = argparse.ArgumentParser(
        description="slyp is a linter and fixer for Python code",
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    parser.add_argument(
        "--list", action="store_true", help="list all error and warning codes"
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="count",
        help="increase output verbosity",
        default=0,
        dest="verbosity",
    )
    # hidden option for quick-and-dirty profiling
    parser.add_argument("--debug-timings", action="store_true", help=argparse.SUPPRESS)
    parser.add_argument(
        "--use-git-ls", action="store_true", help="find python files from git-ls-files"
    )
    parser.add_argument(
        "--disable",
        help="Disable error and warning codes (comma delimited)",
        default="",
    )
    parser.add_argument(
        "--enable",
        help=(
            "Enable error and warning codes which are otherwise disabled "
            "(comma delimited, overrides --disable)"
        ),
        default="",
    )
    parser.add_argument(
        "--no-cache",
        help="Disable caching of results in the '.slyp_cache' directory.",
        action="store_true",
    )
    parser.add_argument("files", nargs="*", help="default: all python files")
    args = parser.parse_args()

    if args.list:
        list_codes()
        sys.exit(0)

    if args.use_git_ls and args.files:
        parser.error("--use-git-ls requires no filenames as arguments")

    # parse inputs from comma delimited lists
    disabled_codes = {x for x in args.disable.split(",") if x != ""}
    enabled_codes = {x for x in args.enable.split(",") if x != ""}
    # add default disables if "all" is not in --enable
    if "all" not in enabled_codes:
        disabled_codes = disabled_codes | DEFAULT_DISABLED_CODES

    success = True
    timings = {}
    if not args.no_cache:
        passing_cache: PassingFileCache | None = PassingFileCache(
            config_id=compute_config_id(enabled_codes, disabled_codes)
        )
    else:
        passing_cache = None

    for filename in all_py_filenames(args.files, args.use_git_ls):
        hashed_file: HashedFile | None
        if passing_cache:
            hashed_file = HashedFile(filename)
            if hashed_file in passing_cache:
                if args.verbosity > 1:
                    print(f"cache hit: {filename}")
                continue
        else:
            hashed_file = None

        start = time.time()
        this_file_success = fix_file(filename, verbose=bool(args.verbosity)) and success
        after_fix = time.time()
        this_file_success = (
            check_file(
                filename,
                verbose=bool(args.verbosity),
                disabled_codes=disabled_codes,
                enabled_codes=enabled_codes,
            )
            and this_file_success
        )
        after_check = time.time()
        timings[filename] = {
            "fix": after_fix - start,
            "check": after_check - after_fix,
        }
        success = this_file_success and success

        if passing_cache and hashed_file:
            if this_file_success:
                passing_cache.add(hashed_file)
            else:
                del passing_cache[hashed_file]

    if args.debug_timings:
        print(json.dumps(timings, indent=2, separators=(",", ": ")))

    if not success:
        sys.exit(1)

    if args.verbosity:
        print("ok")


def all_py_filenames(files: t.Sequence[str], use_git_ls: bool) -> t.Iterable[str]:
    if files:
        yield from files
    elif use_git_ls:
        git_ls_files_proc = subprocess.run(
            ["git", "ls-files"], check=True, capture_output=True, text=True
        )
        candidates = git_ls_files_proc.stdout.split("\n")
        for file in candidates:
            if file == "":
                continue
            if is_python(file):
                yield file
    else:
        yield from glob.glob("**/*.py", recursive=True)


def is_python(filename: str) -> bool:
    # don't pay attention to symlinks or other special filetypes
    # those aren't really files to check
    # but you might be fed them by `git ls-files`
    if not stat.S_ISREG(os.lstat(filename).st_mode):
        return False

    # .py is good
    if filename.endswith(".py"):
        return True

    # if it's not executable, it couldn't be a script
    if not os.access(filename, os.X_OK):
        return False

    # but otherwise, look for a shebang which contains 'python' as a substring
    try:
        with open(filename, encoding="utf-8") as fp:
            firstline = fp.readline()
    except UnicodeDecodeError:
        return False
    if firstline.startswith("#!") and "python" in firstline:
        return True

    return False


def list_codes() -> None:
    first = True
    for code in CODE_MAP.values():
        if code.hidden:
            continue
        if first:
            first = False
        else:
            print()

        if code.default_disabled:
            description: str = f"{code.message} (disabled by default)"
        else:
            description = code.message
        print(f"{code.code}: {description}")
        print(textwrap.indent(code.example, "    "))


def compute_config_id(enabled_codes: set[str], disabled_codes: set[str]) -> str:
    # this is the base config ID for caching
    # it should be updated if fixers are changed or linters are modified
    # in order to ensure that we bust the cache when we ship new rules
    base = "_v1_"

    # now we get the codes which are defined, convert to a string
    all_codes: str = json.dumps(sorted(CODE_MAP.keys()))
    # get the enabled/disabled codes, and make that a string
    code_opts = json.dumps(
        {
            "disabled": sorted(enabled_codes),
            "enabled": sorted(disabled_codes),
        }
    )

    config_hash = hashlib.sha256()
    config_hash.update(all_codes.encode())
    config_hash.update(code_opts.encode())

    # full ID is the base + the computed bits hashed
    return base + config_hash.hexdigest()



import textwrap

import pytest

from slyp.fixer import fix_file


@pytest.fixture
def fix_text(tmpdir):
    def _fix_text(
        text, *, expect_changes=True, verbose=False, dedent=True, filename="X.py"
    ):
        handle = tmpdir.join(filename)
        text = textwrap.dedent(text) if dedent else text
        handle.write(text)
        with tmpdir.as_cwd():
            res = fix_file(filename, verbose=verbose)
            new_text = handle.read()

            if expect_changes:
                assert not res, "fixing made no changes"
            else:
                assert res, "fixing made changes"
                assert text == new_text

            return new_text

    return _fix_text



from __future__ import annotations

import dataclasses
import hashlib
import os
import shutil

_CACHEDIR = ".slyp_cache"


@dataclasses.dataclass
class HashedFile:
    filename: str
    sha: bytes = b""

    def __post_init__(self) -> None:
        if not self.sha:
            self.sha = self._compute_sha()

    def _compute_sha(self) -> bytes:
        with open(self.filename, "rb") as fp:
            return hashlib.sha256(fp.read()).digest()


def _ensure_cachedir(base_cache_dir: str, cache_subdirs: list[str]) -> None:
    os.makedirs(base_cache_dir, exist_ok=True)
    for subdir in cache_subdirs:
        os.makedirs(os.path.join(base_cache_dir, subdir), exist_ok=True)
    gitignore_path = os.path.join(base_cache_dir, ".gitignore")
    if not os.path.exists(gitignore_path):
        with open(gitignore_path, "wb") as fp:
            fp.write(b"*\n")


class PassingFileCache:
    def __init__(
        self,
        *,
        config_id: str,
        base_cache_dir: str = _CACHEDIR,
        cache_dir: str = "passing_files",
    ):
        self._cache_dir = os.path.join(base_cache_dir, cache_dir, config_id)
        _ensure_cachedir(base_cache_dir, [cache_dir])

    def clear(self) -> None:
        shutil.rmtree(_CACHEDIR, ignore_errors=True)

    def __contains__(self, item: HashedFile) -> bool:
        cached_hashfile = os.path.join(self._cache_dir, item.filename)
        if not os.path.exists(cached_hashfile):
            return False
        with open(cached_hashfile, "rb") as fp:
            return fp.read() == item.sha

    def add(self, item: HashedFile) -> None:
        target_file = os.path.join(self._cache_dir, item.filename)
        os.makedirs(os.path.dirname(target_file), exist_ok=True)
        with open(target_file, "wb") as fp:
            fp.write(item.sha)

    def __delitem__(self, item: HashedFile) -> None:
        cached_hashfile = os.path.join(self._cache_dir, item.filename)
        if os.path.exists(cached_hashfile):
            os.remove(cached_hashfile)



from __future__ import annotations

import typing as t

import libcst
import libcst.matchers

# SimpleWhitespace defines whitespace as seen between tokens in most contexts
# it can contain newlines *if* there is a preceding backslash escape
SIMPLE_WHITESPACE_NO_NEWLINE_MATCHER = libcst.matchers.SimpleWhitespace(
    value=libcst.matchers.MatchIfTrue(lambda x: "\n" not in x)
)

PARENT_NODE_REQUIRES_INNERMOST_PARENS_MATCHER = libcst.matchers.OneOf(
    libcst.matchers.Arg(star=libcst.matchers.DoNotCare()),
    libcst.matchers.UnaryOperation(),
)

ParenFixNodeTypes = t.Union[
    libcst.Name,
    libcst.Attribute,
    libcst.Subscript,
    libcst.Call,
    libcst.Dict,
    libcst.DictComp,
    libcst.List,
    libcst.ListComp,
    libcst.Set,
    libcst.SetComp,
    libcst.Tuple,
    libcst.UnaryOperation,
    libcst.BinaryOperation,
    libcst.Comparison,
    libcst.Ellipsis,
    libcst.Integer,
    libcst.Float,
    libcst.SimpleString,
    libcst.ConcatenatedString,
    libcst.FormattedString,
    libcst.Lambda,
    libcst.Yield,
    libcst.GeneratorExp,
    libcst.IfExp,
    libcst.Await,
]
PFN = t.TypeVar("PFN", bound=ParenFixNodeTypes)


class SlypTransformer(libcst.CSTTransformer):
    METADATA_DEPENDENCIES = (
        libcst.metadata.PositionProvider,
        libcst.metadata.ParentNodeProvider,
    )

    def _singular_parens_are_same_line(
        self, node: libcst.With | libcst.ImportFrom
    ) -> bool:
        if node.lpar is not None and libcst.matchers.matches(
            node.lpar, libcst.matchers.LeftParen()
        ):
            lpar_line = self.get_metadata(
                libcst.metadata.PositionProvider, node.lpar  # type: ignore[arg-type]
            ).start.line
            rpar_line = self.get_metadata(
                libcst.metadata.PositionProvider, node.rpar  # type: ignore[arg-type]
            ).start.line
            return lpar_line == rpar_line

        return False

    def _how_many_parens_same_line(self, node: PFN) -> int:
        """
        Determine how many parens are on the same line for a node.

        Start by finding the "max offset" into the paren lists where the lpar and rpar
        are on the same line. The range is [-1 , len(paren_list)] ; then add one.
        """
        max_offset = -1
        for offset in range(len(node.lpar)):
            lpar_line = self.get_metadata(
                libcst.metadata.PositionProvider, node.lpar[-(offset + 1)]
            ).start.line
            rpar_line = self.get_metadata(
                libcst.metadata.PositionProvider, node.rpar[offset]
            ).start.line

            if lpar_line != rpar_line:
                break
            max_offset = offset
        return max_offset + 1

    def modify_parenthesized_node(
        self, original_node: PFN, updated_node: PFN, *, preserve_innermost: bool = False
    ) -> PFN:
        num_parens_to_unwrap: int | None = None
        if not preserve_innermost:
            # **optimization**
            #
            # if there is a parent node to check, pick up on the number of parens to
            # unwrap first -- this saves the work of accessing the parent node when it's
            # avoidable and we can short-circuit
            #
            # based on timings, accessing parent nodes seems slower than finding paren
            # linenos
            num_parens_to_unwrap = self._how_many_parens_same_line(original_node)
            if num_parens_to_unwrap <= 0:
                return updated_node

            parent = self.get_metadata(
                libcst.metadata.ParentNodeProvider, original_node
            )
            # check if the parent of the node is *-expansion of an arg
            if libcst.matchers.matches(
                parent, PARENT_NODE_REQUIRES_INNERMOST_PARENS_MATCHER
            ):
                preserve_innermost = True

        if preserve_innermost and len(original_node.lpar) == 1:
            return updated_node

        if num_parens_to_unwrap is None:
            num_parens_to_unwrap = self._how_many_parens_same_line(original_node)
        if preserve_innermost:
            num_parens_to_unwrap -= 1
        if num_parens_to_unwrap <= 0:
            return updated_node
        return updated_node.with_changes(  # type: ignore[return-value]
            lpar=updated_node.lpar[:-num_parens_to_unwrap],
            rpar=updated_node.rpar[num_parens_to_unwrap:],
        )

    # identifiers, attributes, lookups, and calls

    def leave_Name(
        self, original_node: libcst.Name, updated_node: libcst.Name
    ) -> libcst.Name:
        if not original_node.lpar:
            return updated_node
        return self.modify_parenthesized_node(original_node, updated_node)

    def leave_Attribute(
        self, original_node: libcst.Attribute, updated_node: libcst.Attribute
    ) -> libcst.Attribute:
        if not original_node.lpar:
            return updated_node
        return self.modify_parenthesized_node(original_node, updated_node)

    def leave_Subscript(
        self, original_node: libcst.Subscript, updated_node: libcst.Subscript
    ) -> libcst.Subscript:
        if not original_node.lpar:
            return updated_node
        return self.modify_parenthesized_node(original_node, updated_node)

    def leave_Call(
        self, original_node: libcst.Call, updated_node: libcst.Call
    ) -> libcst.Call:
        if not original_node.lpar:
            return updated_node
        return self.modify_parenthesized_node(original_node, updated_node)

    # collections

    def leave_Dict(
        self, original_node: libcst.Dict, updated_node: libcst.Dict
    ) -> libcst.Dict:
        if not original_node.lpar:
            return updated_node
        return self.modify_parenthesized_node(original_node, updated_node)

    def leave_DictComp(
        self, original_node: libcst.DictComp, updated_node: libcst.DictComp
    ) -> libcst.DictComp:
        if not original_node.lpar:
            return updated_node
        return self.modify_parenthesized_node(original_node, updated_node)

    def leave_List(
        self, original_node: libcst.List, updated_node: libcst.List
    ) -> libcst.List:
        if not original_node.lpar:
            return updated_node
        return self.modify_parenthesized_node(original_node, updated_node)

    def leave_ListComp(
        self, original_node: libcst.ListComp, updated_node: libcst.ListComp
    ) -> libcst.ListComp:
        if not original_node.lpar:
            return updated_node
        return self.modify_parenthesized_node(original_node, updated_node)

    def leave_Set(
        self, original_node: libcst.Set, updated_node: libcst.Set
    ) -> libcst.Set:
        if not original_node.lpar:
            return updated_node
        return self.modify_parenthesized_node(original_node, updated_node)

    def leave_SetComp(
        self, original_node: libcst.SetComp, updated_node: libcst.SetComp
    ) -> libcst.SetComp:
        if not original_node.lpar:
            return updated_node
        return self.modify_parenthesized_node(original_node, updated_node)

    def leave_Tuple(
        self, original_node: libcst.Tuple, updated_node: libcst.Tuple
    ) -> libcst.Tuple:
        # list of tuples:
        #   [(1, 2), (3, 4)] != [1, 2, 3, 4]
        if len(original_node.lpar) < 2:
            return updated_node
        return self.modify_parenthesized_node(
            original_node,
            updated_node,
            preserve_innermost=True,
        )

    # operators

    def leave_UnaryOperation(
        self, original_node: libcst.UnaryOperation, updated_node: libcst.UnaryOperation
    ) -> libcst.UnaryOperation:
        # parens define precedence
        if len(original_node.lpar) < 2:
            return updated_node
        return self.modify_parenthesized_node(
            original_node,
            updated_node,
            preserve_innermost=True,
        )

    def leave_BinaryOperation(
        self,
        original_node: libcst.BinaryOperation,
        updated_node: libcst.BinaryOperation,
    ) -> libcst.BinaryOperation:
        # parens define precedence
        if len(original_node.lpar) < 2:
            return updated_node
        return self.modify_parenthesized_node(
            original_node,
            updated_node,
            preserve_innermost=True,
        )

    def leave_Comparison(
        self, original_node: libcst.Comparison, updated_node: libcst.Comparison
    ) -> libcst.Comparison:
        # parens define precedence
        if len(original_node.lpar) < 2:
            return updated_node
        return self.modify_parenthesized_node(
            original_node,
            updated_node,
            preserve_innermost=True,
        )

    # literals, constants, and strings

    def leave_Ellipsis(
        self, original_node: libcst.Ellipsis, updated_node: libcst.Ellipsis
    ) -> libcst.Ellipsis:
        if not original_node.lpar:
            return updated_node
        return self.modify_parenthesized_node(original_node, updated_node)

    def leave_Integer(
        self,
        original_node: libcst.Integer,
        updated_node: libcst.Integer,
    ) -> libcst.Integer:
        # direct access to attributes of integers and floats requires parens:
        #   (1).bit_length()
        if len(original_node.lpar) < 2:
            return updated_node
        return self.modify_parenthesized_node(
            original_node,
            updated_node,
            preserve_innermost=True,
        )

    def leave_Float(
        self,
        original_node: libcst.Float,
        updated_node: libcst.Float,
    ) -> libcst.Float:
        # direct access to attributes of integers and floats requires parens:
        #   (1).bit_length()
        if len(original_node.lpar) < 2:
            return updated_node
        return self.modify_parenthesized_node(
            original_node,
            updated_node,
            preserve_innermost=True,
        )

    def leave_SimpleString(
        self, original_node: libcst.SimpleString, updated_node: libcst.SimpleString
    ) -> libcst.SimpleString:
        if not original_node.lpar:
            return updated_node
        return self.modify_parenthesized_node(original_node, updated_node)

    def leave_ConcatenatedString(
        self,
        original_node: libcst.ConcatenatedString,
        updated_node: libcst.ConcatenatedString,
    ) -> libcst.ConcatenatedString | libcst.SimpleString:
        new_node = updated_node
        # if the node is parenthesized, potentially strip parens
        if original_node.lpar:
            new_node = self.modify_parenthesized_node(original_node, updated_node)

        # if the node is a pair of simple strings with no newline between them,
        # this may be a chance to join them into a single string node
        if libcst.matchers.matches(
            new_node,
            libcst.matchers.ConcatenatedString(
                whitespace_between=SIMPLE_WHITESPACE_NO_NEWLINE_MATCHER,
                left=libcst.matchers.SimpleString(),
                right=libcst.matchers.SimpleString(),
            ),
        ):
            # check that the left and right match in their prefix and quote characters,
            # and forbid this change (for now) when one of the two is a multiline string
            # (TODO: consider when it might be safe to join multiline strings)
            #
            # the restriction against differing quote characters allows us to avoid any
            # attempt to manipulate unescaped quotes in the string values
            # having done this verification, join the strings into a single node,
            # preserving the prefix and quote style
            left: libcst.SimpleString = new_node.left  # type: ignore[assignment]
            right: libcst.SimpleString = new_node.right  # type: ignore[assignment]
            if (
                (
                    left.prefix == right.prefix
                    or {left.prefix, right.prefix}.issubset({"br", "rb"})
                )
                and left.quote == right.quote
                and left.quote in {"'", '"'}
            ):
                return libcst.SimpleString(
                    lpar=new_node.lpar,
                    rpar=new_node.rpar,
                    value=left.prefix
                    + left.quote
                    + left.raw_value
                    + right.raw_value
                    + left.quote,
                )

        return new_node

    def leave_FormattedString(
        self,
        original_node: libcst.FormattedString,
        updated_node: libcst.FormattedString,
    ) -> libcst.FormattedString:
        if not original_node.lpar:
            return updated_node
        return self.modify_parenthesized_node(original_node, updated_node)

    # expressions

    def leave_Lambda(
        self, original_node: libcst.Lambda, updated_node: libcst.Lambda
    ) -> libcst.Lambda:
        # lambdas can have attributes:
        #   (lambda x: x + 1).__code__
        # and can be called (although it's sort of pointless):
        #   (lambda x: x + 1)(2)
        # and also may parse (for a human) less ambiguously with parens:
        #   (lambda x: x + 1), 2
        if len(original_node.lpar) < 2:
            return updated_node
        return self.modify_parenthesized_node(
            original_node,
            updated_node,
            preserve_innermost=True,
        )

    def leave_Yield(
        self, original_node: libcst.Yield, updated_node: libcst.Yield
    ) -> libcst.Yield:
        new_node = updated_node
        # inject whitespace if a yield is used without trailing space
        # e.g. `yield('foo')` -> `yield ('foo')`
        if new_node.value is not None and (
            new_node.whitespace_after_yield.empty  # type: ignore[union-attr]
        ):
            new_node = new_node.with_changes(
                whitespace_after_yield=libcst.SimpleWhitespace(" ")
            )
        # expression nesting:
        #   return (yield from inner())
        if len(new_node.lpar) < 2:
            return new_node
        return self.modify_parenthesized_node(
            original_node,
            new_node,
            preserve_innermost=True,
        )

    def leave_GeneratorExp(
        self, original_node: libcst.GeneratorExp, updated_node: libcst.GeneratorExp
    ) -> libcst.GeneratorExp:
        # generator expression passed to a function call:
        #   foo((x % 2 for x in range(10)), bar)
        if len(original_node.lpar) < 2:
            return updated_node
        return self.modify_parenthesized_node(
            original_node,
            updated_node,
            preserve_innermost=True,
        )

    def leave_IfExp(
        self, original_node: libcst.IfExp, updated_node: libcst.IfExp
    ) -> libcst.IfExp:
        # parens are required for usage of the results:
        #   (foo() if x else bar())["baz"]
        if len(original_node.lpar) < 2:
            return updated_node
        return self.modify_parenthesized_node(
            original_node,
            updated_node,
            preserve_innermost=True,
        )

    def leave_Await(
        self, original_node: libcst.Await, updated_node: libcst.Await
    ) -> libcst.Await:
        # parens are required for usage of the results:
        #   (await foo())["bar"]
        if len(original_node.lpar) < 2:
            return updated_node
        return self.modify_parenthesized_node(
            original_node,
            updated_node,
            preserve_innermost=True,
        )

    def leave_Match(
        self, original_node: libcst.Match, updated_node: libcst.Match
    ) -> libcst.Match:
        # inject whitespace if missing
        # ensures that `match(x): ...` converts to `match x: ...`
        if not original_node.whitespace_after_match.empty:
            return updated_node
        return updated_node.with_changes(
            whitespace_after_match=libcst.SimpleWhitespace(" ")
        )

    def leave_With(
        self, original_node: libcst.With, updated_node: libcst.With
    ) -> libcst.With:
        # remove parens from `with (foo): ...` as long as they are on the same line
        changes: dict[str, t.Any] = {}
        if original_node.whitespace_after_with.empty:
            changes["whitespace_after_with"] = libcst.SimpleWhitespace(" ")
        if self._singular_parens_are_same_line(original_node):
            changes["lpar"] = libcst.MaybeSentinel.DEFAULT
            changes["rpar"] = libcst.MaybeSentinel.DEFAULT
        if changes:
            return updated_node.with_changes(**changes)
        return updated_node

    def leave_If(self, original_node: libcst.If, updated_node: libcst.If) -> libcst.If:
        # inject whitespace if missing
        # ensures that `if(x): ...` converts to `if x: ...`
        if not original_node.whitespace_before_test.empty:
            return updated_node
        return updated_node.with_changes(
            whitespace_before_test=libcst.SimpleWhitespace(" ")
        )

    def leave_ImportFrom(
        self, original_node: libcst.ImportFrom, updated_node: libcst.ImportFrom
    ) -> libcst.ImportFrom:
        changes: dict[str, t.Any] = {}
        if original_node.whitespace_after_import.empty:
            changes["whitespace_after_import"] = libcst.SimpleWhitespace(" ")
        if self._singular_parens_are_same_line(original_node):
            changes["lpar"] = None
            changes["rpar"] = None
        if changes:
            return updated_node.with_changes(**changes)
        return updated_node
